package uk.ac.ed.inf.mandelbrotmaps;

/**
    Complex implements a complex number and defines complex
    arithmetic and mathematical functions
    Last Updated February 27, 2001
    Copyright 1997-2001
    @version 1.0
    @author Andrew G. Bennett
*/
public class Complex extends Object {

    private double x,y;
    
    /**
        Constructs the complex number z = u + i*v
        @param u Real part
        @param v Imaginary part
    */
    public Complex(double u,double v) {
        x=u;
        y=v;
    }
    
    /**
        Real part of this Complex number 
        (the x-coordinate in rectangular coordinates).
        @return Re[z] where z is this Complex number.
    */
    public double real() {
        return x;
    }
    
    /**
        Imaginary part of this Complex number 
        (the y-coordinate in rectangular coordinates).
        @return Im[z] where z is this Complex number.
    */
    public double imag() {
        return y;
    }
    
    /**
        Modulus of this Complex number
        (the distance from the origin in polar coordinates).
        @return |z| where z is this Complex number.
    */
    public double mod() {
        if (x!=0 || y!=0) {
            return Math.sqrt(x*x+y*y);
        } else {
            return 0d;
        }
    }
    
    /**
        Argument of this Complex number 
        (the angle in radians with the x-axis in polar coordinates).
        @return arg(z) where z is this Complex number.
    */
    public double arg() {
        return Math.atan2(y,x);
    }
    
    /**
        Complex conjugate of this Complex number
        (the conjugate of x+i*y is x-i*y).
        @return z-bar where z is this Complex number.
    */
    public Complex conj() {
        return new Complex(x,-y);
    }
    
    /**
        Addition of Complex numbers (doesn't change this Complex number).
        <br>(x+i*y) + (s+i*t) = (x+s)+i*(y+t).
        @param w is the number to add.
        @return z+w where z is this Complex number.
    */
    public Complex plus(Complex w) {
        return new Complex(x+w.real(),y+w.imag());
    }
    
    /**
        Subtraction of Complex numbers (doesn't change this Complex number).
        <br>(x+i*y) - (s+i*t) = (x-s)+i*(y-t).
        @param w is the number to subtract.
        @return z-w where z is this Complex number.
    */
    public Complex minus(Complex w) {
        return new Complex(x-w.real(),y-w.imag());
    }
    
    /**
        Complex multiplication (doesn't change this Complex number).
        @param w is the number to multiply by.
        @return z*w where z is this Complex number.
    */
    public Complex times(Complex w) {
        return new Complex(x*w.real()-y*w.imag(),x*w.imag()+y*w.real());
    }
    
    /**
        Division of Complex numbers (doesn't change this Complex number).
        <br>(x+i*y)/(s+i*t) = ((x*s+y*t) + i*(y*s-y*t)) / (s^2+t^2)
        @param w is the number to divide by
        @return new Complex number z/w where z is this Complex number  
    */
    public Complex div(Complex w) {
        double den=Math.pow(w.mod(),2);
        return new Complex((x*w.real()+y*w.imag())/den,(y*w.real()-x*w.imag())/den);
    }
    
    /**
        Complex exponential (doesn't change this Complex number).
        @return exp(z) where z is this Complex number.
    */
    public Complex exp() {
        return new Complex(Math.exp(x)*Math.cos(y),Math.exp(x)*Math.sin(y));
    }
    
    /**
        Principal branch of the Complex logarithm of this Complex number.
        (doesn't change this Complex number).
        The principal branch is the branch with -pi < arg <= pi.
        @return log(z) where z is this Complex number.
    */
    public Complex log() {
        return new Complex(Math.log(this.mod()),this.arg());
    }
    
    /**
        Complex square root (doesn't change this complex number).
        Computes the principal branch of the square root, which 
        is the value with 0 <= arg < pi.
        @return sqrt(z) where z is this Complex number.
    */
    public Complex sqrt() {
        double r=Math.sqrt(this.mod());
        double theta=this.arg()/2;
        return new Complex(r*Math.cos(theta),r*Math.sin(theta));
    }
    
    // Real cosh function (used to compute complex trig functions)
    private double cosh(double theta) {
        return (Math.exp(theta)+Math.exp(-theta))/2;
    }
    
    // Real sinh function (used to compute complex trig functions)
    private double sinh(double theta) {
        return (Math.exp(theta)-Math.exp(-theta))/2;
    }
    
    /**
        Sine of this Complex number (doesn't change this Complex number).
        <br>sin(z) = (exp(i*z)-exp(-i*z))/(2*i).
        @return sin(z) where z is this Complex number.
    */
    public Complex sin() {
        return new Complex(cosh(y)*Math.sin(x),sinh(y)*Math.cos(x));
    }
    
    /**
        Cosine of this Complex number (doesn't change this Complex number).
        <br>cos(z) = (exp(i*z)+exp(-i*z))/ 2.
        @return cos(z) where z is this Complex number.
    */
    public Complex cos() {
        //return new Complex(Math.cosh(y)*Math.cos(x),-Math.sinh(y)*Math.sin(x));
        return new Complex(cosh(y)*Math.cos(x),-sinh(y)*Math.sin(x));
    }
    
    /**
        Hyperbolic sine of this Complex number 
        (doesn't change this Complex number).
        <br>sinh(z) = (exp(z)-exp(-z))/2.
        @return sinh(z) where z is this Complex number.
    */
    public Complex sinh() {
        return new Complex(sinh(x)*Math.cos(y),cosh(x)*Math.sin(y));
    }
    
    /**
        Hyperbolic cosine of this Complex number 
        (doesn't change this Complex number).
        <br>cosh(z) = (exp(z) + exp(-z)) / 2.
        @return cosh(z) where z is this Complex number.
    */
    public Complex cosh() {
        return new Complex(cosh(x)*Math.cos(y),sinh(x)*Math.sin(y));
    }
    
    /**
        Tangent of this Complex number (doesn't change this Complex number).
        <br>tan(z) = sin(z)/cos(z).
        @return tan(z) where z is this Complex number.
    */
    public Complex tan() {
        return (this.sin()).div(this.cos());
    }
    
    /**
        Negative of this complex number (chs stands for change sign). 
        This produces a new Complex number and doesn't change 
        this Complex number.
        <br>-(x+i*y) = -x-i*y.
        @return -z where z is this Complex number.
    */
    public Complex chs() {
        return new Complex(-x,-y);
    }
    
    /**
        String representation of this Complex number.
        @return x+i*y, x-i*y, x, or i*y as appropriate.
    */
    public String toString() {
        if (x!=0 && y>0) {
            return x+" + "+y+"i";
        }
        if (x!=0 && y<0) {
            return x+" - "+(-y)+"i";
        }
        if (y==0) {
            return String.valueOf(x);
        }
        if (x==0) {
            return y+"i";
        }
        // shouldn't get here (unless Inf or NaN)
        return x+" + i*"+y;
        
    }
    public Complex gamma() {
		double xx,yy,tx,ty,LGx,LGy,sx,sy,ssx,ssy,xi,yi,xo,yo,r;
		double pi,fac,twopi;
		int k,g;
		double d1,d2,e1,e2,f1,f2;
		
		double c[] = {0, 1.000000000000000174663,
			5716.400188274341379136,
			-14815.30426768413909044,
			14291.49277657478554025,
			-6348.160217641458813289,
			1301.608286058321874105,
			-108.1767053514369634679,
			2.605696505611755827729,
			-0.7423452510201416151527e-2,
			0.5384136432509564062961e-7,
			-0.4023533141268236372067e-8};
		
		//x=re;
		//y=im;
		pi=3.141592653589793;
		twopi=pi+pi;
		g=9;
		sx=0;
		sy=0;
		ssx=0;
		ssy=0;
		xx=x;
		yy=y;
		
		if ((Math.round(xx)==xx)&&(yy==0)&&(xx<=0))
		{
			f1=1.79769313486231570e+308d;
			f2=0;
		}
		if ((Math.round(xx)==xx)&&(yy==0)&&(xx>=0))
		{	
			fac=1;
			for (k=1;k<=xx-1;k++)
				fac=fac*k;			
			f1=fac;
			f2=0;			
        }
		if ((Math.round(xx)!=xx)||(yy!=0))
			if (x<0)
			{
				x=-x;
				y=-y;
			}		
		tx=x+g;
		ty=y;
		for (k=g+2;k>=2;k--)
		{	
			r=tx*tx+ty*ty;
			sx=sx+c[k]*tx/r;
			sy=sy-c[k]*ty/r;
			tx=tx-1;
		}
		sx=sx+c[1];   
		ssx=x+g-0.5;
		ssy=y;
		d1=Math.log(Math.sqrt(sx*sx*twopi+sy*sy*twopi));
		d2=Math.atan2(sy,sx);
		e1=Math.log(Math.sqrt(ssx*ssx+ssy*ssy));
		e2=Math.atan2(ssy,ssx);
		xi=e1*(x-0.5)-e2*y;
		yi=e2*(x-0.5)+e1*y;
		LGx=d1+xi-ssx;
		LGy=d2+yi-ssy;
		LGx=Math.exp(LGx);
		f1=LGx*Math.cos(LGy);
		f2=LGx*Math.sin(LGy);
		if (xx<0)
		{
			xi=pi*x;
			yi=pi*y;
			d1=Math.cosh(yi)*Math.sin(xi);
			d2=Math.sinh(yi)*Math.cos(xi);
			xi=x*f1-y*f2;
			yi=x*f2+y*f1;
			xo=xi*d1-yi*d2;
			yo=xi*d2+yi*d1;
			r=xo*xo+yo*yo;
			f1=-pi*xo/r;
			f2=pi*yo/r;
		}
		return new Complex(f1,f2);		
	}
	
	public Complex zeta() {
		double xx,xi,yi,xj,yj,xo,yo,r;
		double hold1,hold2,hold21,hold22,f1,f2;
		int steps,sw,j;
		double mymod;
		double split=0.5;
		double pi=3.141592653589793;
		int myzt=100;
		double tesc=0.001;
		//x=re;
		//y=im;
		xx=x;
		if (x>10)
		{
			f1=1;
			f2=0;
			return new Complex(f1,f2);
		}
		else
		{
			if (x<split)
			{
				x=1-x;
				y=-y;
			}
			steps=myzt;
			f1=0;
			f2=0;
			sw=-1;
			for (j=1;j<=steps;j++)
			{
				sw=-sw;
				
				//hold2=j^(x+i*y);
				xi=Math.log(j);
				xj=x*xi;
				yj=y*xi;
				r=Math.exp(xj);
				hold21=r*Math.cos(yj);
				hold22=r*Math.sin(yj);                
				
				if ((Double.isInfinite(hold21))||(Double.isInfinite(hold22)))
				{
					hold1=0;
					hold2=0;
				}
				else
				{
					r=hold21*hold21+hold22*hold22;
					hold1=sw*hold21/r;
					hold2=-sw*hold22/r;
				}
				f1=f1+hold1;
				f2=f2+hold2;
				mymod=Math.sqrt(hold1*hold1+hold2*hold2);            
				if (mymod/(Math.sqrt(f1*f1+f2*f2))<tesc)
					break;
			}
			sw=-sw;
			
			//hold2=(1+j)^(x+i*y);
			xi=Math.log(1+j);
			xj=x*xi;
			yj=y*xi;
			r=Math.exp(xj);
			hold21=r*Math.cos(yj);
			hold22=r*Math.sin(yj);
			
			if ((Double.isInfinite(hold21))||(Double.isInfinite(hold22)))
			{
				hold1=0;
				hold2=0;
			}
			else
			{
				r=hold21*hold21+hold22*hold22;
				hold1=sw*hold21/r;
				hold2=-sw*hold22/r;
			}
			f1=f1+hold1/2;
			f2=f2+hold2/2;
			
			//hold2=2^(1-x-iy);
			xi=Math.log(2);
			xj=(1-x)*xi;
			yj=-y*xi;
			r=Math.exp(xj);
			hold21=r*Math.cos(yj);
			hold22=r*Math.sin(yj);
			
			if ((Double.isInfinite(hold21))||(Double.isInfinite(hold22)))
			{
				hold1=0;
				hold2=0;
			}
			else
			{
				r=(1-hold21)*(1-hold21)+hold22*hold22;  
				hold1=(1-hold21)/r;
				hold2=hold22/r; 
			}
			hold21=f1*hold1-f2*hold2;
			hold22=f1*hold2+f2*hold1;   
			f1=hold21;
			f2=hold22;
			if(xx<split)
			{
				//hold=(2*pi)^(-z);
				xi=Math.log(2*pi);
				xj=-x*xi;
				yj=-y*xi;
				r=Math.exp(xj);
				hold1=r*Math.cos(yj);
				hold2=r*Math.sin(yj);
				
				//hold2=cos(z*pi/2);
				xi=x*pi/2;
				yi=y*pi/2;
				hold21=Math.cosh(yi)*Math.cos(xi);
				hold22=-Math.sinh(yi)*Math.sin(xi);
				
				//zi=(2*pi)^(-z)*cos(z*pi/2)
				xi=hold1*hold21-hold2*hold22;
				yi=hold1*hold22+hold2*hold21;
				
				Complex z=new Complex(x,y);
				z=z.gamma();
				hold1=z.real();
				hold2=z.imag();
				
				//zo=gamma(z)*f
				xo=hold1*f1-hold2*f2;
				yo=hold1*f2+hold2*f1;
				
				//f=2*zo*zi
				f1=2*(xi*xo-yi*yo);
				f2=2*(xi*yo+yi*xo);
			}
			
			return new Complex(f1,f2);
		}
	}
}


